#
# Copyright (c) 2013-2018 Balabit
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

'''
Scan some class or module object and decorate it's functions and methods with
the supplied decorator function.

Usage:

>>> class MyClass(object):
...     def mymethod(self):
...         return 0
>>> def log(func):
...     def __wrapper(*args, **kwargs):
...         print('Called function %s' % func.__name__)
...         return func(*args, **kwargs)
...     return __wrapper
>>> decorate_module(MyClass, decorator=log)
>>> obj = MyClass()
>>> print(obj.mymethod())
Called function mymethod
0
'''

import inspect
import warnings


class ModuleDecorator(object):
    '''
    Decorate a module automatically with the supplied decorator.
    This is just a helper class for the :func:`decorate` module function.

    The `decorator` argument is the decorator to be applied to functions.
    '''

    def __init__(self, decorator):
        self.__decorator = decorator

    def decorate(self, module):
        if inspect.isclass(module):
            self.__decorate_class(module, use_dict=module.__dict__)

        else:
            self.__decorate_module(module, use_dict=module.__dict__)

    def __is_attribute_mutable(self, object_dict, attr):
        if '__slots__' not in object_dict:
            return True

        return attr in object_dict['__slots__']

    def __decorate_module(self, module, *, use_dict):
        for key, value in self.__iterate_decorables(use_dict):
            if self.__is_object_external(module, value):
                continue

            self.__decorate_item(module, key, value)

    def __is_object_external(self, module, obj):
        return hasattr(obj, '__module__') and obj.__module__ != module.__name__

    def __decorate_class(self, cls, *, use_dict):
        for key, value in self.__iterate_decorables(use_dict):
            self.__decorate_item(cls, key, value)

    def __iterate_decorables(self, use_dict):
        for key, value in use_dict.items():
            if not self.__is_attribute_mutable(use_dict, key):
                continue

            yield key, value

    def __decorate_item(self, module, key, value):
        if inspect.isfunction(value):
            self.__decorate_function(module, key, value)

        elif inspect.isclass(value):
            self.__decorate_class(value, use_dict=value.__dict__)

        elif isinstance(value, property):
            self.__decorate_property(module, key, value)

        elif isinstance(value, (staticmethod, classmethod)):
            self.__decorate_special_method(module, key, value)

    def __decorate_function(self, module, key, value):
        self.__set_attribute(module, key, self.__decorator(value))

    def __decorate_property(self, module, key, value):
        fget = None
        fset = None
        fdel = None

        if value.fget is not None:
            fget = self.__decorator(value.fget)

        if value.fset is not None:
            fset = self.__decorator(value.fset)

        if value.fdel is not None:
            fdel = self.__decorator(value.fdel)

        self.__set_attribute(
            module, key, property(fget=fget, fset=fset, fdel=fdel)
        )

    def __decorate_special_method(self, module, key, value):
        func = self.__decorator(value.__func__)
        decorated = value.__class__(func)

        self.__set_attribute(module, key, decorated)

    def __submodule_of(self, basemodule, submodule):
        return submodule.startswith(basemodule + '.')

    def __set_attribute(self, module, key, value):
        try:
            setattr(module, key, value)

        # We want to catch all errors here since any problems with setting
        # a module attribute to the decorated function is non-fatal.
        # pylint: disable=W0702
        except:  # noqa
            if hasattr(module, '__module__'):
                name = '{}.{}'.format(module.__module__, module.__name__)

            else:
                name = module.__name__

            warnings.warn(
                'Could not decorate {}.{}'.format(name, key),
                category=RuntimeWarning
            )


def decorate_module(module, *, decorator):
    ModuleDecorator(decorator).decorate(module)


__all__ = ['decorate_module']



#
# Copyright (c) 2013-2018 Balabit
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import re
import os.path
import nose

import typesafety


# Monkey-patching the _exc_info_to_string is the only simple way to influence
# how the exception gets formatted. We can't really replace the object with a
# subclass (we use this in a nose plugin, which is one of the many classes
# modifying the TestResult class).
class ExceptionStringConverter:  # pylint: disable=W0212
    TRACE_LINE_PATTERN = re.compile(r'^  File "([^"]*)"')

    @classmethod
    def wrap_results_object(cls, result):
        result._exc_info_to_string = cls(result._exc_info_to_string)

    def __init__(self, original_converter):
        self.__original_converter = original_converter

    def __call__(self, err, test):
        res = []
        skip_line = False
        for line in self.__original_converter(err, test).split('\n'):
            if skip_line:
                skip_line = False
                continue

            match = self.TRACE_LINE_PATTERN.match(line)
            if match is not None and self.__should_skip(match.group(1)):
                skip_line = True
                continue

            res.append(line)

        return '\n'.join(res)

    def __should_skip(self, filename):
        relative = os.path.relpath(filename, typesafety.__path__[0])
        return not relative.startswith('..')


class TypesafetyPlugin(nose.plugins.Plugin):
    name = 'typesafety'
    __enabled_for = ()
    keep_typesafety_trace = False
    enabled = False

    def __init__(self, *args, activate=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.__activate = activate or typesafety.activate

    def options(self, parser, env):
        parser.add_option(
            '-T', '--enable-typesafety', action='append', metavar='MODULE',
            help='Enable typesafety for the given modules'
        )
        parser.add_option(
            '--keep-typesafety-trace', action='store_true',
            help='Do not hide typesafety traceback frames ' +
            '(useful when debugging typesafety)'
        )

    def configure(self, options, conf):
        if options.enable_typesafety:
            self.enabled = True
            self.__enabled_for = tuple(
                mod.split('.') for mod in options.enable_typesafety
            )
            try:
                self.__activate(filter_func=self.__check_need_activate)

            except RuntimeError:
                # Nose plugin was already enabled in a different thread
                return

            self.keep_typesafety_trace = options.keep_typesafety_trace

    # This is an interface function that cannot be removed (see the nose
    # plugin documentation for the meaning of this function).
    def prepareTestResult(self, result):  # pylint: disable=C0103
        if not self.keep_typesafety_trace:
            ExceptionStringConverter.wrap_results_object(result)

    def __check_need_activate(self, module_name):
        module_name = module_name.split('.')
        return any(
            module_name[:len(name)] == name for name in self.__enabled_for
        )



#
# Copyright (c) 2013-2018 Balabit
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import functools
import inspect
import warnings

from typing_inspect import is_union_type, get_args


class TypesafetyError(Exception):
    '''
    A type error detected by the Typesafety tool. Does not interfere
    with the builtin TypeError so one does not accidentally write an
    assertion in a unit test on a Typesafety error when intending to
    assert on a TypeError raised by actual production code.
    '''


class Validator(object):
    '''
    A Validator is a class that can check the function argument
    and return value types as specified in the function annotations.

    The `function` argument must be a function, method or generator.

    The given function and it's annotations will be checked if they
    conform to the following rules:

    * The annotation is a class (subclass of `type`),
    * the annotation is a callable object (`callable` returns true on it) or
    * is a tuple whose elements conform to the same rules.

    Any annotations not conforming to the above rules will be ignored
    as they might belong to a different purpose.
    '''

    ARG_TYPE_ERROR_MESSAGE = "Argument {0} of function {1!r} is invalid " + \
                             "(expected: {2}; got: {3})"
    RET_TYPE_ERROR_MESSAGE = "Return value of function {0!r} is invalid " + \
                             "(expected: {1}; got: {2})"

    @classmethod
    def get_function_validator(cls, function):
        '''
        Return the validator bound to `function` or None if the function
        is not validated.
        '''

        return getattr(function, '__validator__', None)

    @classmethod
    def is_function_validated(cls, function):
        '''
        Return True if the function has a bound validator.
        '''

        return cls.get_function_validator(function) is not None

    @classmethod
    def decorate(cls, function):
        '''
        Decorate a function so the function call is checked whenever
        a call is made. The calls that do not need any checks are skipped.

        The `function` argument is the function to be decorated.

        The return value will be either

        * the function itself, if there is nothing to validate, or
        * a proxy function that will execute the validation.
        '''

        should_skip = getattr(function, 'typesafety_skip', False)
        if cls.is_function_validated(function) or should_skip:
            return function

        validator = cls(function)

        if not validator.need_validate_arguments and \
                not validator.need_validate_return_value:
            return function

        @functools.wraps(function)
        def __wrapper(*args, **kwargs):
            return validator(*args, **kwargs)

        __wrapper.__validator__ = validator

        return __wrapper

    @classmethod
    def undecorate(cls, function):
        '''
        Remove validator decoration from a function.

        The `function` argument is the function to be cleaned from
        the validator decorator.
        '''

        if cls.is_function_validated(function):
            return cls.get_function_validator(function).function

        return function

    def __init__(self, function):
        self.__function = function
        self.__spec = inspect.getfullargspec(function)
        self.__argument_annotation = {}
        self.__return_annotation = None
        self.__defaults = {}

        self.__process_type_annotations()
        self.__process_return_value_annotation()
        self.__process_default_values()

    @property
    def need_validate_arguments(self):
        '''
        True if any of the function arguments need to be checked.
        '''

        return bool(self.__argument_annotation)

    @property
    def need_validate_return_value(self):
        '''
        True if the return value of the function needs to be be checked.
        '''

        return self.__return_annotation is not None

    @property
    def function(self):
        '''
        The raw function value.
        '''

        return self.__function

    def validate_arguments(self, locals_dict):
        '''
        Validate the arguments passed to a function. If an error occurred,
        the function will throw a TypesafetyError.

        The `locals_dict` argument should be the local value dictionary of
        the function. An example call would be like:
        '''

        for key, value, validator in self.__map_arguments(locals_dict):
            if not self.__is_valid(value, validator):
                key_name = repr(key)
                func_name = self.__function.__name__
                annotation = self.__argument_annotation.get(key)
                message = self.ARG_TYPE_ERROR_MESSAGE.format(
                    key_name,
                    func_name,
                    self.__format_expectation(annotation),
                    value.__class__.__name__)
                raise TypesafetyError(message)

    def __format_expectation(self, annotation):
        if isinstance(annotation, tuple):
            return "({})".format(
                ", ".join(self.__format_expectation(a) for a in annotation)
            )

        if annotation is None:
            return "None"

        return getattr(annotation, '__name__', str(annotation))

    def validate_return_value(self, retval):
        '''
        Validate the return value of a function call. If an error occurred,
        the function will throw a TypesafetyError.

        The `retval` should contain the return value of the function call.
        '''

        if self.__return_annotation is None:
            return

        if not self.__is_valid(retval, self.__return_annotation):
            func_name = self.__function.__name__
            msg = self.RET_TYPE_ERROR_MESSAGE.format(
                func_name,
                self.__format_expectation(self.__return_annotation),
                retval.__class__.__name__
            )
            raise TypesafetyError(msg)

    def __call__(self, *args, **kwargs):
        '''
        Proxy function to the function call including the validations.
        '''

        locals_dict = self.__collect_argument_dictionary(args, kwargs)
        self.validate_arguments(locals_dict)

        # The function property is callable, but pylint sees it as a
        # simple property object.
        return_value = self.function(*args, **kwargs)  # pylint: disable=E1102

        self.validate_return_value(return_value)
        return return_value

    def __process_type_annotations(self):
        for name, value in self.__spec.annotations.items():
            if name == 'return' or \
                    not self.__is_valid_typecheck_annotation(value):
                continue

            self.__argument_annotation[name] = value

    def __process_return_value_annotation(self):
        if 'return' in self.__spec.annotations:
            return_annotation = self.__spec.annotations['return']
            if self.__is_valid_typecheck_annotation(return_annotation):
                self.__return_annotation = return_annotation

    def __process_default_values(self):
        if self.__spec.defaults is not None:
            args_with_defaults = self.__spec.args[-len(self.__spec.defaults):]
            for index, name in enumerate(args_with_defaults):
                self.__defaults[name] = self.__spec.defaults[index]

        if self.__spec.kwonlydefaults is not None:
            for name in self.__spec.kwonlyargs:
                if name in self.__spec.kwonlydefaults:
                    self.__defaults[name] = self.__spec.kwonlydefaults[name]

    def __collect_argument_dictionary(self, args, kwargs):
        locals_dict = dict(self.__defaults)

        for index, value in enumerate(args):
            if index < len(self.__spec.args):
                locals_dict[self.__spec.args[index]] = value

        for name, value in kwargs.items():
            locals_dict[name] = value

        return locals_dict

    def __map_arguments(self, locals_dict):
        for name in self.__spec.args + self.__spec.kwonlyargs:
            if name not in self.__argument_annotation:
                continue

            if name not in locals_dict:
                msg = 'Missing required argument {!r}'.format(name)
                raise TypesafetyError(msg)

            annotation = self.__argument_annotation[name]
            yield name, locals_dict[name], annotation

    def __is_valid(self, value, validator):
        if isinstance(validator, tuple):
            return any(
                self.__is_valid(value, subvalidator)
                for subvalidator in validator
            )

        if is_union_type(validator):
            return any(
                self.__is_valid(value, subvalidator)
                for subvalidator in get_args(validator)
            )

        if isinstance(validator, type):
            return isinstance(value, validator)

        if hasattr(validator, '__call__'):
            return validator(value)

        if validator is None:
            return value is None

        # This line will probably never be reached
        return True

    def __is_valid_typecheck_annotation(self, validator):
        if isinstance(validator, tuple):
            is_valid = all(
                self.__is_valid_typecheck_annotation(subvalidator)
                for subvalidator in validator
            )
            if is_valid:
                warnings.warn("Tuple notation is deprecated, use typing.Union or typing.Optional", DeprecationWarning)

            return is_valid

        if is_union_type(validator):
            return all(
                self.__is_valid_typecheck_annotation(subvalidator)
                for subvalidator in get_args(validator)
            )

        if isinstance(validator, type):
            return True

        if callable(validator):
            if validator == callable:
                warnings.warn(
                    "Using callable() as a notation is deprecated, use typing.Callable instead",
                    DeprecationWarning
                )
            return True

        if validator is None:
            return True

        return False


__tracebackhide__ = True
__all__ = ['Validator']



#
# Copyright (c) 2013-2018 Balabit
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import collections
import inspect


# This is a sphinx interface implementation so the number of
# arguments is outside our control.
def add_annotations_to_signature(
        app,
        what,
        name,
        obj,
        options,
        signature,
        return_annotation
):  # pylint: disable=R0913
    if what not in {"function", "method", "class"}:
        return None

    if what == "class":
        obj = obj.__init__
        what = "method"

        if not inspect.isfunction(obj):
            return None

    func = __extract_decorated_function(obj)
    arg_spec = inspect.getfullargspec(func)

    if not arg_spec.annotations:
        return None

    return __format_signature(what, func, arg_spec)


def __should_stop_resolution(obj, depth, *, maxdepth=100):
    return depth >= maxdepth or \
           not hasattr(obj, "decorated_function") or \
           not inspect.isfunction(obj.decorated_function)


def __extract_decorated_function(obj):
    depth = 0

    while not __should_stop_resolution(obj, depth):
        obj = obj.decorated_function
        depth += 1

    return obj


def __format_signature(type_str, function, arg_spec):
    return_type = __format_return_type(arg_spec)
    args = __format_args_and_kwargs(type_str, function, arg_spec)

    return (args, return_type)


def __format_return_type(arg_spec):
    annotations = arg_spec.annotations

    if "return" not in annotations:
        return ""

    return __format_type_annotation(annotations["return"])


def __format_type_annotation(annotation):
    if hasattr(annotation, "__name__"):
        result = annotation.__name__
        module = getattr(annotation, '__module__', 'builtins')
        if module != "builtins":
            result = module + '.' + result

        return result

    if isinstance(annotation, str):
        return repr(annotation)

    if isinstance(annotation, collections.Sequence):
        annotation_sequence = ", ".join(
            __format_type_annotation(t) for t in annotation
        )
        return "(" + annotation_sequence + ")"

    return str(annotation)


def __format_args_and_kwargs(type_str, function, arg_spec):
    args = __format_positional_args(type_str, function, arg_spec)
    annotations = arg_spec.annotations

    if arg_spec.varargs is not None:
        args.append("*%s" % __format_arg(arg_spec.varargs, annotations, {}))

    if arg_spec.kwonlyargs:
        args.append("*")
        args.extend(
            __format_args(
                arg_spec.kwonlyargs,
                annotations,
                arg_spec.kwonlydefaults
            )
        )

    if arg_spec.varkw is not None:
        args.append("**%s" % __format_arg(arg_spec.varkw, annotations, {}))

    return "(%s)" % ", ".join(args)


def __format_positional_args(type_str, function, arg_spec):
    arg_names, defaults = __remove_self_or_cls(type_str, function, arg_spec)
    defaults_by_name = __get_defaults_as_dict(arg_names, defaults)

    return __format_args(arg_names, arg_spec.annotations, defaults_by_name)


def __remove_self_or_cls(type_str, function, arg_spec):
    arg_names = arg_spec.args
    defaults = arg_spec.defaults

    if type_str == "method" or inspect.ismethod(function):
        if defaults is not None and len(defaults) == len(arg_names):
            defaults = defaults[1:]

        arg_names = arg_names[1:]

    return arg_names, defaults


def __get_defaults_as_dict(arg_names, defaults):
    defaults = defaults if defaults is not None else ()
    offset = len(arg_names) - len(defaults)

    return dict(zip(arg_names[offset:], defaults))


def __format_args(arg_names, annotations, defaults_by_name):
    args = []

    for arg_name in arg_names:
        args.append(__format_arg(arg_name, annotations, defaults_by_name))

    return args


def __format_arg(arg_name, annotations, defaults_by_name):
    name = __format_argument_name(arg_name, annotations)
    default_value_str = __format_default_value(arg_name, defaults_by_name)

    return "%s%s" % (name, default_value_str)


def __format_argument_name(arg_name, annotations):
    if arg_name not in annotations:
        return arg_name

    return arg_name + ": " + __format_type_annotation(annotations[arg_name])


def __format_default_value(arg_name, defaults_by_name):
    if not defaults_by_name or arg_name not in defaults_by_name:
        return ""

    return "=%s" % repr(defaults_by_name[arg_name])


def setup(app):
    app.connect("autodoc-process-signature", add_annotations_to_signature)



#
# Copyright (c) 2013-2018 Balabit
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

'''
On import decorate imported modules with the supplied decorator.
'''

import imp
import importlib.abc
import sys
from . import autodecorator


class ModuleLoader(object):
    '''
    Simple proxy object that will call the :func:`ModuleFinder.load_module`
    function with the proper parameters.
    '''

    def __init__(self, finder, fullname, path):
        self.__finder = finder
        self.__fullname = fullname
        if path is None:
            self.__name = fullname

        else:
            self.__name = fullname.rsplit('.', 1)[-1]

        self.__path = path
        self.__info = imp.find_module(self.__name, path)

    @property
    def fullname(self):
        return self.__fullname

    @property
    def name(self):
        return self.__name

    @property
    def info(self):
        return self.__info

    def load_module(self, fullname):
        '''
        Load the module. Required for the Python meta-loading mechanism.
        '''

        return self.__finder.load_module(self)


class ModuleFinder(object):
    '''
    Module finder and loader. When installed, modules will be
    automatically decorated with the supplied decorator.

    The `decorator` is the decorator function to apply.

    The `filter` argument is a filter function that should return
    True if the given module should be decorated. This function takes
    two arguments:

    * The full name of the module and
    * the module object itself (after import).
    '''

    __decorator = None
    __filter = None
    __loaded_modules = None

    def __init__(self, decorator):
        self.__decorator = decorator
        self.__reset()

    @property
    def installed(self):
        '''
        True if the module finder/loader has been installed.
        '''

        return self in sys.meta_path

    def set_filter(self, filter_func):
        '''
        Set the module filter function.

        The `filter_func` argument expects a callable that gets the full
        module name and the loaded module object.
        '''

        self.__filter = filter_func

    def install(self):
        '''
        Install the module finder. If already installed, this will do nothing.
        After installation, each loaded module will be decorated.
        '''

        if not self.installed:
            sys.meta_path.insert(0, self)

    def uninstall(self):
        '''
        Uninstall the module finder. If not installed, this will do nothing.
        After uninstallation, none of the newly loaded modules will be
        decorated (that is, everything will be back to normal).
        '''

        if self.installed:
            sys.meta_path.remove(self)

        # Reload all decorated items
        import_list = []
        for name in self.__loaded_modules:
            del sys.modules[name]
            import_list.append(name)

        for name in import_list:
            __import__(name)

        self.__reset()

    def find_module(self, fullname, path=None):
        '''
        Find the module. Required for the Python meta-loading mechanism.

        This will do nothing, since we use the system to locate a module.
        '''

        loader = None
        if self.__filter is None or self.__filter(fullname):
            loader = ModuleLoader(self, fullname, path)

        return loader

    def load_module(self, loader):
        '''
        Load the module. Required for the Python meta-loading mechanism.
        '''

        modfile, pathname, description = loader.info
        module = imp.load_module(
            loader.fullname,
            modfile,
            pathname,
            description
        )
        sys.modules[loader.fullname] = module
        self.__loaded_modules.add(loader.fullname)

        autodecorator.decorate_module(module, decorator=self.__decorator)

        return module

    def __reset(self):
        self.__loaded_modules = set()


importlib.abc.Loader.register(ModuleLoader)
importlib.abc.Finder.register(ModuleFinder)


__all__ = ['ModuleFinder']



#
# Copyright (c) 2013-2018 Balabit
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

'''
Type safety checker module for python. The type safety checker uses
a debugger to check each function call and return and check the
value types according to the annotations specified.

Accepted annotations for type checked functions are class objects,
callable objects or tuples containing class and callable objects. If
any of the annotations does not conform to this rule the annotation
will be ignored.
'''

from .validator import Validator, TypesafetyError
from .finder import ModuleFinder


class Typesafety(object):
    '''
    Singleton class for managing the type safety checker.
    '''

    __instance = None

    @classmethod
    def instance(cls):
        '''
        Return the singleton instance.
        '''

        if cls.__instance is None:
            cls.__instance = cls()

        return cls.__instance

    def __init__(self):
        self.__module_finder = None

    @property
    def active(self):
        '''
        Returns whether the type safety checker is enabled.
        '''

        return self.__module_finder is not None

    def activate(self, *, filter_func=None):
        '''
        Activate the type safety checker. After the call all functions
        that need to be checked will be.
        '''

        if self.active:
            raise RuntimeError("Type safety check already active")

        self.__module_finder = ModuleFinder(Validator.decorate)
        if filter_func is not None:
            self.__module_finder.set_filter(filter_func)
        self.__module_finder.install()

    def deactivate(self):
        '''
        Deactivate the type safety checker. After the call no functions
        will be checked.
        '''

        if not self.active:
            raise RuntimeError("Type safety check inactive")

        self.__module_finder.uninstall()


def activate(*, filter_func=None):
    '''
    Shorthand function for activating the type checking.
    '''

    Typesafety.instance().activate(filter_func=filter_func)


def deactivate():
    '''
    Shorthand function for deactivating the type checking.
    '''

    Typesafety.instance().deactivate()


__all__ = ['Typesafety', 'TypesafetyError', 'activate', 'deactivate']



#
# Copyright (c) 2013-2018 Balabit
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
from typesafety import TypesafetyError


def pytest_addoption(parser):
    parser.addoption(
        '-T', '--enable-typesafety', action='append', metavar='MODULE',
        help='Enable typesafety for the given modules'
    )


def pytest_configure(config):
    if config.getoption('enable_typesafety'):
        import typesafety
        import functools

        enabled_for = tuple(
            mod.split('.') for mod in config.getoption('enable_typesafety')
        )
        filter_func = functools.partial(__check_need_activate,
                                        enabled_for=enabled_for)
        typesafety.activate(filter_func=filter_func)


def pytest_unconfigure(config):
    if config.getoption('enable_typesafety'):
        import typesafety
        typesafety.deactivate()


def pytest_runtest_makereport(item, call):
    """
    Use report maker hook for early traceback cleanup

    Have to remove typesafety from trace before py.test removes itself, for
    avoid empty trace.
    """

    excinfo = call.excinfo
    if not excinfo or excinfo.type is not TypesafetyError:
        return

    excinfo.traceback = excinfo.traceback.filter()


def __check_need_activate(module_name, enabled_for):
    module_name = module_name.split('.')
    return any(
        module_name[:len(name)] == name for name in enabled_for
    )



#!/usr/bin/env python3
#
# Copyright (c) 2013-2018 Balabit
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

try:
    from setuptools import setup

except ImportError:
    from distutils.core import setup

setup(
    name="typesafety",
    description="Type safety checker for Python3",
    long_description= """\
Typesafety is a tool for writing type-checked code in Python. In languages
like C++, Java, etc., this is a language-level feature, but Python has no
such feature. With the advent of annotations it is however possible to write
code with type notations. Typesafety is a means to enforce that those notations
are valid.
""",
    license="LGPLv2+",
    version="2.1.0",
    author="Viktor Hercinger",
    author_email="viktor.hercinger@balabit.com",
    maintainer="Viktor Hercinger",
    maintainer_email="viktor.hercinger@balabit.com",
    keywords='nose type typesafe static',
    url="https://github.com/balabit/typesafety",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Software Development :: Testing',
        'Topic :: Documentation :: Sphinx',
    ],
    zip_safe=True,
    use_2to3=False,
    packages=['typesafety'],
    entry_points={
        'nose.plugins.0.10': [
            'typesafety = typesafety.noseplugin:TypesafetyPlugin'
        ],
        'pytest11': ['pytest_typesafety = typesafety.pytest_typesafety'],
    },
    requires=[
        'nose',
        'sphinx',
        'typing_inspect',
    ]
)
